#!/usr/bin/python
from __future__ import with_statement
import random, re, sys
random.seed(5738)

morse={
    'A':'.-',
    'B':'-...',
    'C':'-.-.',
    'D':'-..',
    'E':'.',
    'F':'..-.',
    'G':'--.',
    'H':'....',
    'I':'..',
    'J':'.---',
    'K':'-.-',
    'L':'.-..',
    'M':'--',
    'N':'-.',
    'O':'---',
    'P':'.--.',
    'Q':'--.-',
    'R':'.-.',
    'S':'...',
    'T':'-',
    'U':'..-',
    'V':'...-',
    'W':'.--',
    'X':'-..-',
    'Y':'-.--',
    'Z':'--..',
    '0':'-----',
    '1':'.----',
    '2':'..---',
    '3':'...--',
    '4':'....-',
    '5':'.....',
    '6':'-....',
    '7':'--...',
    '8':'---..',
    '9':'----.',
    '.':'.-.-.-',
    ',':'--..--',
    '?':'..--..',
    "'":'.----.',
    '!':'-.-.--',
    '/':'-..-.',
    '(':'-.--.',
    ')':'-.--.-',
    '&':'.-...',
    ':':'---...',
    ';':'-.-.-.',
    '=':'-...-',
    '+':'.-.-.',
    '-':'-....-',
    '_':'..--.-',
    '"':'.-..-.',
    '$':'...-..-',
    '@':'.--.-.',
    '*':'......', # map * to "ERROR" prosign
    #'\xd3': '---.', # o with an accent, also o-slash and o-with-two dots
}
# make reverse mapping dictionary
revmorse=dict((v,k) for k,v in morse.iteritems())
# add lower-case alpha to morse
for k,v in morse.items():
    morse[k.lower()] = v

def to_morse(w):
    return ' '.join(morse[c] for c in w)
def from_morse(m):
    return ''.join(revmorse.get(l, '*') for l in m.split())
def mlen(w):
    return len(to_morse(w))

def flipm(m):
    return m.replace('.','X').replace('-','.').replace('X','-')
def reversed_morse(w):
    return from_morse(''.join(reversed(to_morse(w))))
def flip_morse(w):
    m = flipm(to_morse(w))
    return from_morse(m)
def reversed_flip_morse(w):
    m = flipm(to_morse(w))
    m = ''.join(reversed(m))
    return from_morse(m)

def test():
    print to_morse("ScottAnanian")
    print mlen("Scott")
    print from_morse(to_morse("Scott"))
    print reversed_morse("Scott")
    print flip_morse("Scott")

def dict_search(filter_func=None, output_func=None):
    words = set()
    while True:
        word = sys.stdin.readline()
        if not word: break
        word = word.strip().replace(' ','')
        if "'" in word: continue
        if not re.match(r'^[A-Za-z]*$', word): continue
        word = word.upper()
        words.add(word)

    def default_filter(word, morse, reversed, flip, reversed_flip):
        #if reversed in words or flip in words or reversed_flip in words:
        #if reversed in words and flip in words and len(word)>3:
        #if (reversed in words or flip in words) and '*' not in reversed and '*' not in flip and len(word)>4 and mlen(word)<32:
        #if True or (flip in words and len(word)>4):
        #if reversed_flip in words and len(word)>4:
        return True
    def default_output(word, morse, reversed, flip, reversed_flip):
        #print word, r, f, rf, mlen(word)
        print word, to_morse(word).replace(' ',''),
        print len(to_morse(word).replace(' ',''))

    if filter_func is None: filter_func = default_filter
    if output_func is None: output_func = default_output

    for word in words:
        m = to_morse(word)
        r = reversed_morse(word)
        f = flip_morse(word)
        rf = reversed_flip_morse(word)
        if filter_func(word, m, r, f, rf):
            output_func(word, m, r, f, rf)

def tob48(word=None, morse=None, align=False):
    """Convert word to our internal 48-bit representation."""
    m = morse if morse else to_morse(word)
    r,cnt = 0, 0
    for c in m:
        if c == '.':
            r = (r<<2) | 3 # dot, then intersymbol space
            cnt += 2
        elif c == '-':
            r = (r<<2) | 1 # dash, then intersymbol space
            cnt += 2
        elif c == ' ':
            r = r ^ 1 # make intersymbol space into interletter
        else:
            assert False
    # remove last intersymbol space
    r = (r>>1)
    cnt -= 1
    # always 48 bits
    if align:
        i = cnt
        while i < 48:
            r = (r << 1) # zero-filled
            i += 1
    return r,cnt

def tob32(word=None, morse=None):
    """Convert word to our internal 32-bit representation (no spaces)."""
    m = morse if morse else to_morse(word)
    r,cnt = 0, 0
    for c in m:
        if c == '.':
            r = (r<<1) | 1 # dot (1)
            cnt += 1
        elif c == '-':
            r = (r<<1) | 0 # dash (0)
            cnt += 1
        elif c == ' ':
            pass # omit spaces
        else:
            assert False
    # end with 10...0
    r = (r<<1) | 1
    # always 32 bits
    assert cnt <= 32
    i = cnt+1 # include the terminating 1 bit
    while i < 32:
        r = (r << 1) # zero-filled
        i += 1
    return r,cnt

def swap4(n):
    return ((n&0xFF)>>4) | ((n<<4)&0xFF)
def com8(n):
    return 0xFF - (n&0xFF)
def ror(n):
    return ((n&0xFF)>>1) | ((n<<7)&0xFF)
def rol(n):
    return ((n&0xFF)>>7) | ((n<<1)&0xFF)
def hash(word=None, morse=None):
    r,cnt = tob48(word, morse, align=True)
    # now fold all the bytes together
    h = 0
    h ^= cnt * 2
    h ^= ((r>> 0)&0xFF) * 1
    h ^= ((r>> 8)&0xFF) * 1
    h ^= ((r>>16)&0xFF) * 1
    h = swap4(h)
    h ^= ((r>>24)&0xFF) * 1
    h = ror(h)
    h ^= ((r>>32)&0xFF) * 1
    h ^= ((r>>40)&0xFF) * 1
    if cnt==19:
        h -= 4
        if ((r>>46)&1) == 0: h+= 1
    h = h & 0xFF
    return h, com8(h)

def hash_answers():
    answers = {
                "JAIHO":      ("MI","JO"), #3S, 15 dit/dahs
                "CARSEAT":    ("CH","HN"), #4S, 16 dit/dahs
                "BYLINE":     ("AE","MA"), #2S, 17 dit/dahs
                "LOVEETC":    ("LO","RT"), #4S, 18 dit/dahs
                "CARPAL":     ("FM","IN"), #2S, 19 dit/dahs
                "THESOUTH":   ("ON","LE"), #2S, 20 dit/dahs
                "GODUTCH":    ("TY","AH"), #3S, 21 dit/dahs
                "SUBSTITUTE": ("PY","YN"), #1S, 22 dit/dahs
                "CASHCOW":    ("TH","OV"), #4S, 23 dit/dahs
                "MARSROVER":  ("ON","EL"), #4S, 24 dit/dahs

                "FOOT":       ("10","10"),                 # 11 sym
                   "IAMB":                  ("IN", "AT"),  # 10 sym
                "TASTE":      ("11","11"),                 #  8 sym
                   "UMAMI":                 ("MA", "LO"),  # 11 sym
                "PARTY":      ("12","12"),                 # 14 sym
                   "WHIG":                  ("RC", "NG"),  # 12 sym
                "DONKEY":     ("13","13"),                 # 16 sym
                   "EEYORE":                ("HN", "HO"),  # 13 sym
                "SANDWICH":   ("14","14"),                 # 23 sym
                   "REUBEN":                ("IN", "RN"),  # 14 sym
                "FABRIC":     ("15","15"),                 # 19 sym
                   "TAFFETA":               ("ET", "SP"),  # 15 sym
                "MECHANICS":  ("16","16"),                 # 24 sym <- longest
                   "FLUID":                 ("YS", "EE"),  # 16 sym
                "HAIRSTYLE":  ("17","17"),                 # 24 sym <- longest
                   "TOPKNOT":               ("EV", "DW"),  # 17 sym
                "SEASON":     ("18","18"),                 # 14 sym
                   "WINTERTIME":            ("EN", "AY"),  # 18 sym
                }
    return answers

def build_gperf_input():
    """Use 'gperf -n -m 1000' on the output."""
    answers = hash_answers()
    print "%compare-lengths"
    print "%null-strings"
    print "%7bit"
    print "%global-table"
    print "%struct-type"
    print "struct output { char *name; char letter; };"
    print "%%"
    def flipabit(m, WHICH=6):
        return m[0:WHICH] + flipm(m[WHICH]) + m[(WHICH+1):]
    def flipbits(m, lst):
        for which in lst:
            m = flipabit(m, which)
        return m
    def tweak(m, orient, which_letter):
        if orient: m = flipm(m)
        m = flipabit(flipabit(m, 7), 3) if which_letter else m
        m = m.replace('-', 'O') # space out dash and dot
        #m = m + 'a' if which_letter else m + 'b'
        return m
    for w, (a1, a2) in answers.iteritems():
        m = to_morse(w).replace(' ','')
        if m == m[::-1]: # palindromes use lower case
            a1, a2 = a1.lower(), a2.lower()
        else:
            a1, a2 = a1.upper(), a2.upper()
        # pad out to uniform length
        if len(m) < 32:
            m += '.' # 1
            while len(m) < 32:
                m += '-' # 0...0
        # orientation 1
        print "#", w
        print "\"%s\", %d" % (tweak(m, False, False), ord(a1[0]))
        print "\"%s\", %d" % (tweak(m, False, True ), ord(a1[1]))
        # orientation 2
        print "\"%s\", %d" % (tweak(m, True,  False), ord(a2[0]))
        print "\"%s\", %d" % (tweak(m, True,  True ), ord(a2[1]))

def build_boring_table():
    answers = hash_answers()
    # add easter egg
    KEY_OFFSET = 11
    eei = easter_egg_info()
    answers[eei['trigger'].replace(' ','')] = ('**','**')
    ee_message = eei['message'] + '\0'
    # output
    print "; Answer mapping. 4 bytes morse -> 2 bytes, 2 bytes (answer letters)"
    print "; Total memory consumption:", 8 * len(answers) + len(ee_message), "bytes"
    print "; Morse: 1=short (dot), 0=long (dash)"
    print "; left aligned, terminated with 10...0"
    lines = []
    for w, (a1, a2) in answers.iteritems():
        r,cnt = tob32(w)
        b = padbin(r, 32)
        s = ", ".join("B'%s'" % b[i:i+8] for i in xrange(0,32,8))
        b = [int(b[i:i+8], 2) for i in xrange(0,32,8)]
        # interleave answer letters
        s += ', "%s%s%s%s"' % (a1[0], a2[0], a1[1], a2[1])
        b += [ord(c) for c in [a1[0], a2[0], a1[1], a2[1]]]
        # comment string
        s = '\tDT\t'+s
        c = '\t; %s -> %s, %s\n\t; %s' % (w, a1, a2, to_morse(w))
        lines.append((s,c,b))
    lines.sort()  # heck, we could do binary search if we wanted to
    print "boring_table"
    allbytes = []
    for s,c,b in lines:
        print c
        print s
        allbytes += b
    print "#define EASTER_EGG_KEY_OFFSET", KEY_OFFSET
    print "easter_egg"
    print "\t;", ee_message[:-1]
    print "\t;", '   '.join(to_morse(w) for w in ee_message[:-1].split())
    for c,b in zip(ee_message, allbytes[KEY_OFFSET:]):
        print "\tDT\tB'%s'" % padbin(ord(c) ^ b, 8)


def build_hash_table():
    SHOW_WORK=False
    letters="ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-"
    table = [ None ] * 256
    answers = hash_answers()
    # now insert the entries we need
    def addone(w, h, a):
        if table[h] is not None and table[h] != a:
            print "conflict adding", w, h, a
            assert False, "conflict adding "+w
        table[h] = a
    for w, (a1, a2, also_reverse) in answers.iteritems():
        h1,h2 = hash(w)
        if SHOW_WORK:
            print w, h1, "->", a1, "/", h2, "->", a2, '\t   [%d]'%tob48(w)[1]
            #print w, padbin(tob48(w, align=True)[0], 48)
        addone(w, h1, a1)
        addone(w, h2, a2)
        if not also_reverse:
            a1, a2 = '*', '*'
        h1, h2 = hash(morse=''.join(reversed(to_morse(w))))
        if SHOW_WORK:
            print w, h1, "->", a1, "/", h2, "->", a2
            #print w, padbin(tob48(morse=''.join(reversed(to_morse(w))), align=True)[0], 48)
        addone(w, h1, a1)
        addone(w, h2, a2)
        h1, h2 = hash(morse=flipm(to_morse(w)))
        if SHOW_WORK:
            print w, h1, "->", a1, "/", h2, "->", a2
            #print w, padbin(tob48(morse=flipm(to_morse(w)), align=True)[0], 48)
        addone(w, h1, a1)
        addone(w, h2, a2)
        h1, h2 = hash(morse=''.join(reversed(flipm(to_morse(w)))))
        if SHOW_WORK:
            print w, h1, "->", a1, "/", h2, "->", a2
            #print w, padbin(tob48(morse=''.join(reversed(flipm(to_morse(w)))), align=True)[0], 48)
        addone(w, h1, a1)
        addone(w, h2, a2)

    table = [ x if x is not None else '*' # random.sample(letters,1)[0]
              for x in table ]
    #print table2
    # print out as assembly table
    print "\tDT\t" + repr(''.join(table)).replace("'",'"')

#MARCH97
#SRLSHOW
# answers: ORATE, SKEWS, QUAKE, FRONT, TSETSE, SENSE, MCKAY

def test2():
    r,cnt=tob48("SCOTTIE")
    print cnt, bin(r)
    h1,h2 = hash("SCOTTIE")
    print bin(h1), bin(h2)
    print to_morse("JOHNADAMS")
    r,cnt=tob48("JOHNADAMS")
    print bin(r),cnt
    for w in sorted(["JAIHO","BYLINE","CARSEAT","LOVEETC","PINHEAD",
                    "GODUTCH","OCTOPOD",
                    #"BUFFOON","CASHCOW",
                    #"RESTSTOP",
                    "FRONT","MCKAY","ORATE","QUAKE","SENSE","SKEWS","TSETSE"
                    ]):
        r,cnt=tob48(w)
        print w, cnt

def padbin(n, width):
    b = bin(n)[2:]
    while len(b) < width:
        b = '0' + b
    return b

def build_morse_table():
    print "; Morse code character table."
    print "; 1=short (dot), 0=long (dash)"
    print "; left aligned, terminated with 10...0"
    for i in xrange(ord(' ')+1, ord('Z')+1):
        c = chr(i)
        if c == '!':
            print "morse_punct"
        if c == 'A':
            print "morse_letter"
        num_bytes = 1 # used to be 2 if c < 'A' else 1
        r,cnt = tob32(c) if c in morse else tob32('*')
        b = padbin(r, 32)
        out = "\tDT\tB'%s'" % b[0:8]
        if num_bytes == 2:
            out += ", B'%s'" % b[8:16]
        print out+"\t;", c, hex(ord(c)), '*' if (c not in morse or c=='*') else morse[c]

def test3():
    r,cnt = tob48("SOS", align=True)
    r = padbin(r, 32)
    print cnt, r[0:8], r[8:16], r[16:24], r[24:]
    h1,h2 = hash("SOS")
    print "SOS", padbin(h1, 9), padbin(h2, 9)

def test4():
    answers = hash_answers()
    for w in sorted(answers.keys()):
        r,cnt = tob32(w)
        print '%10s' % w, padbin(r, 32), '%2d' % cnt, to_morse(w)

def easter_egg_info():
    return {
        "trigger": "OUT OF CASH",
        "message": "PROCEED TO 5-234",
        }

def make_easter_egg(message_obscure):
    KEY_OFFSET = 11
    eei = easter_egg_info()
    trigger_obscure = 'FISH'
    trigger = eei['trigger']
    #message_obscure= "e1euoyunrefxds1do"
    message = eei['message'] + '\0'

    print "\t;; Easter egg trigger for end game."
    print "\t;; Trigger is", trigger, "in our morse representation"
    print "\t;;", '   '.join(to_morse(w) for w in trigger.split())
    r,cnt = tob32(trigger.replace(' ',''))
    pb = padbin(r,32)
    i = 3
    for x, bits in zip(trigger_obscure, [ pb[0:8], pb[8:16],
                                          pb[16:24], pb[24:32] ]):
        xx,_ = tob32(x)
        nb = int(bits, 2) ^ int(padbin(xx,32)[0:8], 2)
        print "#define EASTER_EGG_K%d A'%s'" % (i, x)
        print "#define EASTER_EGG_D%d B'%s'" % (i, padbin(nb, 8))
        i -= 1
    print "#define EASTER_EGG_KEY_OFFSET", KEY_OFFSET
    print "easter_egg"
    for x, m in zip(message_obscure[KEY_OFFSET:], message):
        nb = ord(x) ^ ord(m)
        print "\tDT\tB'%s'" % padbin(nb, 8), ";", repr(chr(nb))

if __name__ == '__main__':
    #test2()
    #build_hash_table()
    #build_morse_table()
    #test3()
    #dict_search()
    test4()
